from typing import (
    Optional
)
from .channels import GuildStageChannel
from .guild import GuildScheduledEvent
from .application import Application
from .user import User
from .abc import BaseChannel
from .partials import PartialGuild

class Invite:
    def __init__(self, data: dict):
        self.code: str = data["code"]
        self.guild: Optional[PartialGuild] = PartialGuild(data["guild"]) or None
        self.channel: BaseChannel = BaseChannel(data["channel"]) 
        self.inviter: Optional[User] = User(data["inviter"]) or None
        self.target_type: int = data["target_type"]
        self.target_user: Optional[User] = User(data["target_user"]) or None
        self.target_application: Optional[Application] = Application(data["target_application"]) or None
        self.approximate_presence_count: Optional[int] = data["approximate_presence_count"] or None
        self.approximate_member_count: Optional[int] = data["approximate_member_count"] or None
        self.expires_at: Optional[str] = data["expires_at"] or None
        self.stage_instance: Optional[GuildStageChannel] = GuildStageChannel(data["stage_instance"]) or None
        self.guild_scheduled_event: Optional[GuildScheduledEvent] = GuildScheduledEvent(data["guild_scheduled_event"]) or None
    # Dabmaster is gonna work on this


from .abc import BaseSlashCommandOption
from typing import (
    Optional,
    List
)

class Section:
    def __init__(self):
        self.commands = {}
        self.events = {}
    
    def event(self, event_name: str):
        def register_event(func):
            self.events[event_name] = func
        return register_event

    def slash_command(self,*, name: str, description: Optional[str], options: List[BaseSlashCommandOption]):
        def register_slash_command(func):
            self.commands[name] = {
                "callback": func,
                "name": name,
                "description": description,
                "options": options
            }
        return register_slash_command


from typing import (
    Optional,
    List
)

class EmbedAuthor:
    def __init__(self, data: dict):
        self.name: str = data["name"]
        self.url: Optional[str] = data["url"] or None
        self.icon_url: Optional[str] = data["icon_url"] or None
        self.proxy_icon_url: Optional[str] = data["proxy_icon_url"] or None

class Embed: # Always wanted to make this class :D
    def __init__(self, data: dict):
        self.data = data
        self.title: Optional[str] = data["title"] or None
        self.type: Optional[str] = data["type"] or None
        self.description: Optional[str] = data["description"] or None
        self.url: Optional[str] = data["url"] or None
        self.timestamp: Optional[str] = data["timestamp"] or None
        self.color: Optional[str] = data["color"] or None
        self.footer: Optional[str] = data["footer"] or None
        self.image: Optional[str] = data["image"] or None
        self.thumbnail: Optional[str] = data["thumbnail"] or None
        self.valid_styles: Optional[str] = data["valid_styles"] or None
        self.provider: Optional[str] = data["provider"] or None
        self.author: Optional[EmbedAuthor] = EmbedAuthor(data["author"]) or None
        self.fields: Optional[List[str]] = data["fields"] or None
        
    def add_field(self, name: str, value: str, inline: bool):
        self.fields.append({"name": name, "value": value, "inline": inline})
        
    def set_title(self, title: str):
        self.title = title
    
    def set_description(self, description: str):
        self.description = description
    
    # Someone else do this icba


from .user import User

class MessageInteraction: # https://discord.com/developers/docs/interactions/receiving-and-responding#message-interaction-object-message-interaction-structure idk
    def __init__(self, data: dict):
        self.id: str = data["id"]
        self.user: User = User(data["user"])
        self.name: str = data["name"]
        self.type: int = data["type"]


from typing import (
    Optional
)

class SlashCommand: # Somethin like Command from commands.ext dpy
    def __init__(self, data: dict):
        self.name: str = data["name"]
        self.description: Optional[str] = data["description"]
        self.options


from .abc import Messageable
from .member import ThreadMember
from .exceptions import ThreadArchived, NotFound404
from .invite import Invite
from .client import Client
from .permissions import Overwrite
from .message import Message
from .partials import PartialUser
from .abc import BaseChannel
from typing import (
    List,
    Optional,
    Dict,
    Union
)

class GuildChannel(BaseChannel):
    def __init__(self, client: Client, data: dict):
        super().__init__(client, data)
        self.guild_id: str = data["guild_id"]
        self.position: int = data["position"]
        self.nsfw: bool = data["nsfw"]
        self.permission_overwrites: List[dict] = data["permission_overwrites"]
        self.parent_id: str = data["parent_id"]
        self.name: str = data["name"]
        
    async def delete(self, *, reason: Optional[str] = None) -> None:
        if reason:
            headers = self.client.http.headers.copy()
            headers["reason"] = reason
            
        response = await self.client.http.delete(f"channels/{self.id}", headers=headers)
        return await response.json()
    
    async def fetch_invites(self) -> List[Invite]: 
        response = await self.client.http.get(f"/channels/{self.id}/invites")
        return await response.json()
    
    async def create_invite(self, *, max_age: Optional[int], max_uses: Optional[int], temporary: Optional[bool], unique: Optional[bool], target_type: Optional[int], target_user_id: Optional[str], target_application_id: Optional[str]):
        data = {}
        data["max_age"] = max_age or None
        data["max_uses"] = max_uses or None
        data["temporary"] = temporary or None
        data["unique"] = unique or None
        data["target_type"] = target_type or None
        data["target_user_id"] = target_user_id or None
        data["target_application_id"] = target_application_id or None
        await self.client.http.post(f"/channels/{self.id}/invites", data=data)   
        
    async def delete_overwrite(self, overwrites: Overwrite) -> None:
        response = await self.client.http.delete(f"/channels/{self.id}/permissions/{overwrites.id}")
        return await response.json()
    
        
    async def fetch_pinned_messages(self) -> List[Message]:
        response = await self.client.http.get(f"/channels/{self.id}/pins")
        data = await response.json()
        return [Message(message) for message in data]
    
    
    
    # async def edit_permission_overwrites I'll do this later
    
    
    # async def edit(self, *,name: str, position: str, permission_overwrites: List[dict], reason: Optional[str] = None):
    #     data = {}
    #     if name:
    #         data["name"] = name
    #     if position:
    #         data["position"] = position
    #     if permission_overwrites:
    #         data["permission_overwrites"] = permission_overwrites
    #     headers = self.client.http.headers
    #     headers["X-Audit-Log-Reason"] = reason
    #     response = await self.client.http.patch(f"channels/{self.id}", data=data, headers=headers)
    #     data = await response.json()
    #     return GuildChannel(self.client, data)
    
class GuildTextChannel(GuildChannel, Messageable):
    def __init__(self, client: Client, data: dict):
        super().__init__(client, data)
        self.topic: str = data["topic"]
        self.rate_limit_per_user: int = data["rate_limit_per_user"]
        self.last_message_id: str = data["last_message_id"]
        self.default_auto_archive_duration: int = data["default_auto_archive_duration"]
    
    async def start_thread(self, name: str,* , auto_archive_duration: Optional[int], type: Optional[int], invitable: Optional[bool], rate_limit_per_user: Optional[int], reason: Optional[str]):
        data = {"name": name}
        if auto_archive_duration:
            data["auto_archive_duration"] = auto_archive_duration
        if type:
            data["type"] = type
        if invitable is not None: # Geez having a bool is gonna be a pain
            data["invitable"] = invitable
        if rate_limit_per_user:
            data["rate_limit_per_user"] = rate_limit_per_user
        
        headers = self.client.http.headers.copy()
        headers["X-Audit-Log-Reason"] = reason
        
        response = await self.client.http.post(f"channels/{self.id}/threads", data=data, headers=headers)
        self.client.guilds[self.guild_id].append(Thread(await response.json()))
    

    
    
    async def bulk_delete(self, message_ids: List[Message.id], reason: Optional[str]) -> None:

        if reason:
            headers = self.client.http.headers.copy()
            headers["X-Audit-Log-Reason"] = reason
            
        response = await self.client.http.post(f"channels/{self.id}/messages/bulk-delete", data={"messages": message_ids}, headers=headers)
        return await response.json()

    async def list_public_archived_threads(self,*, before: Optional[str], limit: Optional[int]) -> Dict[str, Union[List[Messageable], List[ThreadMember], bool]]: # It returns a List of Threads but I can't typehint that...
        response = await self.client.http.get(f"/channels/{self.id}/threads/archived/public",params={"before": before, "limit": limit})
        return await response.json()
    
    async def list_private_archived_threads(self,*, before: Optional[str], limit: Optional[int]) -> Dict[str, Union[List[Messageable], List[ThreadMember], bool]]: # It returns a List of Threads but I can't typehint that...
        response = await self.client.http.get(f"/channels/{self.id}/threads/archived/private",params={"before": before, "limit": limit})
        return await response.json()
    
    async def list_joined_private_archived_threads(self,*, before: Optional[str], limit: Optional[int]) -> Dict[str, Union[List[Messageable], List[ThreadMember], bool]]:
        response = await self.client.http.get(f"/channels/{self.id}/threads/archived/private",params={"before": before, "limit": limit})
        return await response.json()
    
    # async def edit(self,*, name: Optional[str], position: Optional[str], permission_overwrites: Optional[List[dict]], reason: Optional[str], topic: Optional[str], nsfw: bool, rate_limit_per_user: Optional[int], parent_id: Optional[int], default_auto_archive_duration: Optional[int]):
    #     data = {}
    #     if name:
    #         data["name"] = name
    #     if position:
    #         data["position"] = position
    #     if permission_overwrites:
    #         data["permission_overwrites"] = permission_overwrites
        
    #     headers = self.client.http.headers
    #     headers["X-Audit-Log-Reason"] = reason
    #     response = await self.client.http.patch(f"channels/{self.id}", data=data, headers=headers)
    #     data = await response.json()
    #     return GuildTextChannel(self.client, data)
        

    
class GuildNewsChannel(GuildTextChannel):
    def __init__(self, client: Client, data: dict):
        super().__init__(client, data)
        self.default_auto_archive_duration: int = data["default_auto_archive_duration"]

    async def follow(self, webhook_channel_id: str):
        response = await self.client.http.post(f"/channels/{self.id}/followers", data={"webhook_channel_id": webhook_channel_id})
        return await response.json()

class GuildVoiceChannel(GuildChannel):
    def __init__(self, client: Client, data: dict):
        super().__init__(client, data)
        self.bitrate: int = data["bitrate"]
        self.user_limit: int = data["user_limit"]
        self.rtc_region: str = data["rtc_region"]
            
class DMChannel(BaseChannel):
    def __init__(self, client: Client, data: dict):
        super().__init__(client, data)
        self.recipient: List[PartialUser] = PartialUser(data["recipient"])

class ChannelCategory(GuildChannel):
    def __init__(self, client: Client, data: dict):
        super().__init__(client, data)

class GuildStoreChannel(GuildChannel):
    def __init__(self, client: Client, data: dict):
        super().__init__(client, data)
        
class Thread(GuildTextChannel):
    def __init__(self, client: Client, data: dict):
        super().__init__(client, data)
        self.owner_id: str = data["owner_id"]
        self.message_count: int = data["message_count"]
        self.member_count: int = data["member_count"]
        self.archived: bool = data["archived"]
        self.auto_archive_duration: int = data["auto_archive_duration"]
        self.archive_timestamp: str = data["archive_timestamp"]
        self.locked: bool = data["locked"]
    
    async def join(self):
        if self.archived:
            raise ThreadArchived("This thread has been archived so it is no longer joinable")
        response = await self.client.http.put(f"/channels/{self.id}/thread-members/@me")
        return await response.json()
    
    async def add_member(self, member_id: ThreadMember.id):
        if self.archived:
            raise ThreadArchived("This thread has been archived so it is no longer joinable")
        
        response = await self.client.http.put(f"/channels/{self.id}/thread-members/{member_id}")
        return await response.json()
    
    async def leave(self):
        if self.archived:
            raise ThreadArchived("This thread has been archived so it is no longer leaveable")
        response = await self.client.http.delete(f"/channels/{self.id}/thread-members/@me")
        return await response.json()
    
    async def remove_member(self, member_id: ThreadMember.id):
        if self.archived:
            raise ThreadArchived("This thread has been archived so it is no longer leaveable")
        
        response = await self.client.http.delete(f"/channels/{self.id}/thread-members/{member_id}")
        return await response.json()
    
    async def fetch_member(self, member_id: ThreadMember.user_id) -> ThreadMember:
        response = await self.client.http.get(f"/channels/{self.id}/thread-members/{member_id}")
        if response.status == 404:
            raise NotFound404("The member you are trying to fetch does not exist")
        return ThreadMember(await response.json())
    
    async def list_members(self) -> List[ThreadMember]:
        response = await self.client.http.get(f"/channels/{self.id}/thread-members")
        return [ThreadMember(member) for member in await response.json()]
            
class GuildNewsThread(Thread, GuildNewsChannel):
    def __init__(self, client: Client, data: dict):
        super().__init__(client, data)

class GuildStageChannel(BaseChannel):
    def __init__(self, client: Client, data: dict):
        super().__init__(client, data)
        self.guild_id: str = data["guild_id"]
        self.channel_id: str = data["channel_id"]
        self.privacy_level: int = data["privacy_level"]
        self.discoverable_disabled: bool = data["discoverable_disabled"]

class TextBasedChannel(BaseChannel):
    def __init__(self, client: Client, data: dict):
        super().__init__(data)
        if self.type == 0:
            return GuildTextChannel(client, data)
        
        elif self.type == 1:
            return DMChannel(data)
        
        elif self.type == 2:
            return GuildVoiceChannel(client, data)
        
        elif self.type == 4:
            return ChannelCategory(client, data)
        
        elif self.type == 5:
            return GuildNewsChannel(client, data)
        
        elif self.type == 6:
            return GuildStoreChannel(client, data)
        
        elif self.type == 10:
            return GuildNewsThread(client, data)
        
        elif self.type == 11 or self.type == 12:
            return Thread(client, data)
        
        elif self.type == 13:
            return GuildStageChannel(client, data)


from .client import Client

class Route:
    def __init__(self, client: Client,url: str):
        self.client = client
        self.base_url = "https://discord.com/api/v9/"
        self.url = f"{self.base_url}{url}"
    
    async def request(self, method: str, *args, **kwargs):
        return await self.client.http.request(method, self.url, *args, **kwargs)


class MentionedChannel:
    def __init__(self, data: dict):
        self.id: str = data["id"]
        self.guild_id: str = data["guild_id"]
        self.type: int = data["type"]
        self.name: str = data["name"]
        



from .client import Client
from .role import Role
from typing import (
    Optional,
    List
)
from .user import User

class ThreadMember:
    def __init__(self, data: dict):
        self.thread_id: str = data["thread_id"]
        self.user_id: str = data["user_id"]
        self.join_timestamp: str = data["join_timestamp"]
        self.flags: int = data["flags"]
        
class GuildMember:
    def __init__(self, client: Client, data: dict):
        self.data = data
        self.client = client
        self.user: Optional[User] = User(data["user"]) or None
        self.nick: Optional[str] = data["nick"] or None
        self.avatar: Optional[str] = data["avatar"] or None
        self.roles: List[Role] = [Role(role) for role in data["roles"]]
        self.joined_at:str = data["joined_at"]
        self.premium_since: Optional[str] = data["premium_since"] or None
        self.deaf: bool = data["deaf"]
        self.mute: bool = data["mute"]
        self.pending: Optional[bool] = data["pending"] or None
        self.permissions: Optional[str] = data["permissions"] or None
        self.communication_disabled_until: Optional[str] = data["communication_disabled_until"] or None
        



from typing import (
    List,
    Optional
)
from .channels import TextBasedChannel
from .client import Client
from .message import Message
from .exceptions import (
    InvalidArgumentType,
    CustomIdIsTooBig
)

class Messageable:
    def __init__(self, client: Client, channel_id: str):
        self.channel_id: str = channel_id
        self.client = client
        
    async def fetch_messages(self,*, around: Optional[str] = None, before: Optional[str] = None, after: Optional[str] = None, limit: Optional[int] = None) -> List[Message]:
        response = await self.client.http.get(f"channels/{self.id}/messages", params={"around": around, "before": before, "after": after, "limit": limit})
        data = await response.json()
        return [Message(message) for message in data]
    
    async def fetch_message(self,*, message_id: str) -> Message:
        response = await self.client.http.get(f"channels/{self.id}/messages/{message_id}")
        data = await response.json()
        return Message(data)

    async def send(self, message_data: dict) -> Message:
        response = await self.client.http.post(f"channels/{self.id}/messages", data=message_data)
        return Message(await response.json())
    
class BaseSlashCommandOption:
    def __init__(self, *, name: str, description: str, required: bool = False):
        self.settings = {
            "name": name,
            "description": description,
            "required": required
        }
        # People shouldn't use this class, this is just a base class for other options
    
class BaseChannel:
    def __init__(self, client, data: dict):
        self.id: str = data["id"]
        self.client: Client = client
        self.type = data["type"]

class BaseComponent:
    def __init__(self):
        self.settings = {}

    def set_custom_id(self, custom_id: str):
        
        if not isinstance(custom_id, str):
            raise InvalidArgumentType("Custom Id must be a string.")
        
        elif len(custom_id) > 100:
            raise CustomIdIsTooBig("Custom Id must be 100 characters or less.")

        self.settings["custom_id"] = custom_id
        
class BaseInteraction:
    def __init__(self, data: dict):
        self.application_id: str = data["application_id"]
        self.channel: TextBasedChannel = TextBasedChannel(data["channel"])


from urllib.parse import quote
from .reaction import Reaction
from .thread import Thread
from .stickers import StickerItem
from .components import MessageSelectMenu, MessageButton
from .interactions import MessageInteraction
from .partials import PartialMember
from .role import Role
from .mentions import (
    MentionedUser,
    MentionedChannel
)    
from .application import Application
from .client import Client
from .webhook import WebhookUser
from .user import User
from .reaction import Reaction
from typing import (
    Optional,
    List,
    Union
)
from .embed import Embed

class MessageActivity:
    def __init__(self, data: dict):
        self.type: int = data["type"]
        self.party_id: Optional[str] = data["party_id"]

class AllowedMention:
    def __init__(self, allowed_mentions: List[str], replied_user: bool, roles: List[str], users: List[str]):
        self.data = {}
        self.data["parse"] = allowed_mentions
        self.data["replied_user"] = replied_user
        self.data["roles"] = roles
        self.data["users"] = users
        return self.data
    

class Message:
    def __init__(self, client: Client, data: dict):
        self.client = client
        self.id: str = data["id"]
        self.channel_id: str = data["channel_id"]
        self.guild_id: Optional[str] = data["guild_id"] or None
        self.webhook_id: Optional[str] = data["webhook_id"] or None
        self.author: Optional[User] if not self.webhook_id else WebhookUser = WebhookUser(data["author"]) if self.webhook_id else User(data["author"])
        self.member: Optional[PartialMember] = PartialMember(data["member"]) if data["member"] else None
        self.content: str = data["content"]
        self.timestamp: str = data["timestamp"]
        self.edited_timestamp: Optional[str] = data["edited_timestamp"] or None
        self.tts: bool = data["tts"]
        self.mention_everyone: bool = data["mention_everyone"]
        self.mentions: Optional[List[MentionedUser]] = [MentionedUser(mention) for mention in data["mentions"]] or None
        self.mention_roles: Optional[List[int]] = data["mention_roles"] or None
        self.mention_channels: Optional[List[MentionedChannel]] = [MentionedChannel(channel) for channel in data["mention_channels"]] or None
        self.embeds: Optional[List[Embed]] = [Embed(embed) for embed in data["embeds"]] or None
        self.reactions: Optional[List[Reaction]] = [Reaction(reaction) for reaction in data["reactions"]] or None
        self.nonce: Optional[Union[int, str]] = data["nonce"] or None
        self.pinned: bool = data["pinned"]
        self.type: int = data["type"]
        self.activity: MessageActivity = MessageActivity(data["activity"])
        self.application: Application = Application(data["application"]) # Despite there being a PartialApplication, Discord don't specify what attributes it has
        self.flags: int = data["flags"]
        self.referenced_message: Optional[Message] = Message(data["referenced_message"]) if data["referenced_message"] else None
        self.interaction: Optional[MessageInteraction] = MessageInteraction(data["interaction"]) if data["interaction"] else None
        self.thread: Thread = Thread(data["thread"]) if data["thread"] else None
        self.components: Optional[List[Union[MessageSelectMenu, MessageButton]]] = [MessageSelectMenu(component) if component["type"] == 1 else MessageButton(component) for component in data["components"]] or None
        self.stickers: Optional[List[StickerItem]] = [StickerItem(sticker) for sticker in data["stickers"]] or None
        
    async def add_reaction(self, emoji: str):
        emoji = quote(emoji)
        response = await self.client.http.put(f"channels/{self.channel_id}/messages/{self.id}/reactions/{emoji}/@me")
        return await response.json()
    
    async def remove_reaction(self, emoji: str, user: Optional[User] = None):
        emoji = quote(emoji)
        if not user:        
            response = await self.client.http.delete(f"channels/{self.channel_id}/messages/{self.id}/reactions/{emoji}/@me")
        else:
            response = await self.client.http.delete(f"channels/{self.channel_id}/messages/{self.id}/reactions/{emoji}/{user.id}")
        return await response.json()
    
    async def fetch_reactions(self,*, after, limit) -> List[Reaction]:
        response = await self.client.http.get(f"channels/{self.channel_id}/messages/{self.id}/reactions?after={after}&limit={limit}")
        return await response.json()
    
    async def delete_all_reactions(self):
        response = await self.client.http.delete(f"channels/{self.channel_id}/messages/{self.id}/reactions")
        return await response.json()
    
    async def delete_reaction_for_emoji(self, emoji: str):
        emoji = quote(emoji)
        response = await self.client.http.delete(f"channels/{self.channel_id}/messages/{self.id}/reactions/{emoji}")
        return await response.json()
    
    async def edit(self, message_data: dict):
        response = await self.client.http.patch(f"channels/{self.channel_id}/messages/{self.id}", data=message_data)
        return await response.json()
    
    async def delete(self):
        response = await self.client.http.delete(f"channels/{self.channel_id}/messages/{self.id}")
        return await response.json()
    
    async def pin(self, *, reason: Optional[str]):
        headers = self.client.http.headers.copy()
        headers["X-Audit-Log-Reason"] = reason
        response = await self.client.http.put(f"channels/{self.channel_id}/pins/{self.id}", headers=headers)
        return await response.json()
    
    async def unpin(self, *, reason: Optional[str]):
        headers = self.client.http.headers.copy()
        headers["X-Audit-Log-Reason"] = reason
        response = await self.client.http.delete(f"channels/{self.channel_id}/pins/{self.id}", headers=headers)
        return await response.json()

    async def start_thread(self, name: str, auto_archive_duration: Optional[int], rate_limit_per_user: Optional[int]):
        response = await self.client.http.post(f"channels/{self.channel_id}/messages/{self.id}/threads", data={"name": name, "auto_archive_duration": auto_archive_duration, "rate_limit_per_user": rate_limit_per_user})
        self.client.guilds[self.guild_id].append(Thread(await response.json())) # Cache it
        return Thread(await response.json())
    
    async def crosspost(self):
        response = await self.client.http.post(f"channels/{self.channel_id}/messages/{self.id}/crosspost")
        return await response.json()